"""
Basic building blocks for feature resolution.
"""

from __future__ import annotations

import dataclasses
import inspect
from functools import cached_property
from typing import Any, Awaitable, Callable, Generic, Type, cast

from trippa.utils.types import T, V


class Info(Generic[V]):
    """
    Info is used during feature resolution to access other features
    or a generic context given by the client.
    """

    def __init__(self, ctx: V, resolve: Callable[[AnyFeature], Awaitable[Any]]) -> None:
        self.ctx = ctx
        self._resolve = resolve

    async def resolve(self, feature: BaseFeature[T]) -> T:
        return await self._resolve(feature)  # type: ignore


@dataclasses.dataclass
class BaseFeature(Generic[T]):  # pylint: disable=too-many-instance-attributes
    """
    BaseFeature represent the definition of a feature.

    Contains all the required data and metadata to describe a feature.
    """

    name: str
    resolver: Callable[[Info[Any]], Awaitable[T] | T]

    def __hash__(self) -> int:
        return self.name.__hash__()

    def __str__(self) -> str:
        return self.name

    @cached_property
    def description(self) -> str:
        return (self.resolver.__doc__ or "").strip()

    @cached_property
    def type(self) -> Type[T]:
        return cast(
            Type[T],
            self._signature.return_annotation,
        )

    @cached_property
    def _signature(self) -> inspect.Signature:
        return inspect.signature(self.resolver, eval_str=True)


AnyFeature = BaseFeature[Any]



"""
Decorators to create features from resolver functions
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Awaitable, Callable, cast, overload

from trippa.exceptions import MissingInputException
from trippa.utils.types import T

from .base import BaseFeature, Info


@dataclass(frozen=True)
class feature:
    """
    Helper decorator to create feature definitions easily.

    Wraps a function and creates a BaseFeature from it.

    Ex:
    ```python
    @feature()
    async def ciao(ctx: Info) -> str:
        first_ = await ctx.resolve(first)
        return f"{first_}ciao"
    ```
    """

    meta: Any = None
    """Any metadata you want to add that might be eventually retrieved."""

    @overload
    def __call__(
        self,
        resolver: Callable[[Info[Any]], Awaitable[T]],
    ) -> BaseFeature[T]:
        ...

    @overload
    def __call__(
        self,
        resolver: Callable[[Info[Any]], T],
    ) -> BaseFeature[T]:
        ...

    def __call__(
        self,
        resolver: Callable[[Info[Any]], Any],
    ) -> BaseFeature[Any]:
        return BaseFeature[Any](name=resolver.__name__, resolver=resolver)


class RaiseOnMissing(Enum):
    TOKEN = ""


@dataclass(frozen=True)
class input_feature:
    default_value: Any | RaiseOnMissing = RaiseOnMissing.TOKEN

    def __call__(
        self,
        resolver: Callable[[], T],
    ) -> BaseFeature[T]:
        feature_name = resolver.__name__

        def raiser(ctx: Info[Any]) -> T:
            if self.default_value != RaiseOnMissing.TOKEN:
                return cast(T, self.default_value)
            raise MissingInputException(feature_name)

        return BaseFeature[T](
            name=feature_name,
            resolver=raiser,
        )



"""Resolution results wrappers, to handle exception in a more controlled way."""


from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic

from trippa.utils.types import T


@dataclass(frozen=True)
class ResultSuccess(Generic[T]):
    """Wraps a feature result value."""

    value: T


@dataclass(frozen=True)
class ResultFailure:
    """Wraps an exception happened during feature resolution."""

    error: Any


FeatureResult = ResultSuccess[T] | ResultFailure



import asyncio
from typing import Awaitable, cast

from .types import T


async def maybe_await(
    value: Awaitable[T] | T,
) -> T:
    if asyncio.iscoroutine(value):
        return await cast(Awaitable[T], value)
    return cast(T, value)



import dataclasses
from dataclasses import dataclass
from typing import Any, Awaitable, cast

from trippa.base import AnyFeature, BaseFeature
from trippa.results import FeatureResult
from trippa.utils.types import T


@dataclass
class EntityTasks:
    _tasks: dict[AnyFeature, Any] = dataclasses.field(default_factory=dict)

    def __getitem__(self, __key: BaseFeature[T]) -> Awaitable[FeatureResult[T]]:
        return cast(
            Awaitable[FeatureResult[T]],
            self._tasks[__key],
        )

    def get(self, __key: BaseFeature[T]) -> Awaitable[FeatureResult[T]] | None:
        return self._tasks.get(__key)

    def __setitem__(
        self,
        __key: BaseFeature[T],
        item: Awaitable[FeatureResult[T]],
    ) -> None:
        self._tasks[__key] = item



class MissingInputException(Exception):
    def __init__(self, input_name: str) -> None:
        super().__init__(
            f"Feature '{input_name}' missing, try adding it to the inputs."
        )



"""
.. include:: ../../README.md
"""

from .base import AnyFeature as AnyFeature
from .base import BaseFeature as BaseFeature
from .base import Info as Info
from .decorator import feature as feature
from .decorator import input_feature as input_feature
from .resolution import Resolution as Resolution
from .resolution import Trippa as Trippa
from .results import FeatureResult as FeatureResult



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import Any, ClassVar, Mapping, Sequence, Type

from typing_extensions import Self

from trippa.dependencies import find_deps
from trippa.tasks import EntityTasks
from trippa.utils.asyncio import maybe_await
from trippa.utils.types import T

from .base import AnyFeature, BaseFeature, Info
from .results import FeatureResult, ResultFailure, ResultSuccess


@dataclass
class Resolution:
    """
    This is the hearth of `trippa`.

    The Resolution (aka resolution context) keeps track of all the computed features
    and determines how to resolve them.

    Spawn a new Resolution any time you want to calculate stuff
    on a new entity/row or just reset the calculations
    """

    features: Mapping[str, AnyFeature]
    """Features definitions available."""
    ctx: Any
    """Context passed to the resolvers."""

    _inputs: dict[str, FeatureResult[Any]] = field(default_factory=dict)
    _tasks: EntityTasks = field(default_factory=EntityTasks)

    def input(self, inputs: Mapping[AnyFeature, Any] | Mapping[str, Any]) -> Self:
        """Sets a mapping of feature values that will be used in other resolutions."""
        # Set results from inputs
        parsed_input: dict[str, FeatureResult[Any]] = {
            f.name if isinstance(f, BaseFeature) else f: ResultSuccess(v)
            for f, v in inputs.items()
        }
        self._inputs = self._inputs | parsed_input
        return self

    async def resolve_many(
        self, features: Sequence[AnyFeature]
    ) -> dict[AnyFeature, Any]:
        """Resolve many features at once."""
        results = await asyncio.gather(*[self.resolve(f) for f in features])
        return {f: results[i] for i, f in enumerate(features)}

    async def safe_resolve_many(
        self, features: Sequence[AnyFeature]
    ) -> dict[AnyFeature, FeatureResult[Any]]:
        """Same as resolve many but using `safe_resolve`."""
        results: Sequence[FeatureResult] = await asyncio.gather(
            *[self.safe_resolve(f) for f in features]
        )
        return {f: results[i] for i, f in enumerate(features)}

    async def resolve(
        self,
        feature: BaseFeature[T],
    ) -> T:
        """Resolves a feature, returning it's value. Throws on Exception"""
        result = await self.safe_resolve(feature)
        if isinstance(result, ResultFailure):
            raise result.error
        return result.value

    async def safe_resolve(self, feature: BaseFeature[T]) -> FeatureResult[T]:
        """
        Resolves a feature, returning a FeatureResult.

        Does not throw on exception, instead returns a ResultFailure wrapping it
        """
        if result := self._inputs.get(feature.name):
            return result

        if not self._tasks.get(feature):
            self._tasks[feature] = asyncio.create_task(self._safe_resolve(feature))

        return await self._tasks[feature]

    async def _safe_resolve(
        self,
        feature: BaseFeature[T],
    ) -> FeatureResult[T]:
        try:
            return ResultSuccess(await self._instance_resolver(feature))
        except Exception as err:
            return ResultFailure(err)

    async def _instance_resolver(self, feature: BaseFeature[T]) -> T:
        return await maybe_await(feature.resolver(self._info()))

    def _info(self) -> Info[Any]:
        return Info(self.ctx, self.resolve)


class Trippa:
    """Application entrypoint, used to generated resolution contexts'"""

    resolution_class: ClassVar[Type[Resolution]] = Resolution

    features: Mapping[str, AnyFeature]

    def __init__(self, features: Sequence[AnyFeature]) -> None:
        self.features = {f.name: f for f in features}

    def find_deps(self, feature: AnyFeature) -> set[AnyFeature]:
        return find_deps(feature, self.features)

    @classmethod
    def from_modules(cls, modules: list[Any]) -> Self:
        return cls(cls._get_features_from_modules(modules))

    @staticmethod
    def _get_features_from_modules(modules: list[Any]) -> list[AnyFeature]:
        return [
            feature
            for module in modules
            for feature in module.__dict__.values()
            if isinstance(feature, BaseFeature)
        ]

    def start(self, ctx: Any = None) -> Resolution:
        """Starts a new resolution context."""
        return Resolution(self.features, ctx)



"""Module dedicate to resolve dependencies of a feature by reading into it's ast"""
from __future__ import annotations

import ast
from ast import Attribute, Call, ImportFrom, Load, Module, Name, NodeVisitor, alias
from functools import lru_cache
from inspect import getsource, getsourcefile
from pathlib import Path
from typing import Any, Mapping

from . import AnyFeature


@lru_cache(maxsize=5)
def get_source_file_ast(path: Path) -> Module:
    source_text = path.read_text("utf-8")
    return ast.parse(source_text)


class DependencyVisitor(NodeVisitor):
    def __init__(self) -> None:
        self.dependencies = set[str]()

    def visit_Call(self, node: Call) -> Any:
        match node:  # noqa
            case Call(
                func=Attribute(
                    value=Name(id="ctx", ctx=Load()), attr="resolve", ctx=Load()
                ),
                args=[Name(id=feature_alias, ctx=Load())],
            ):
                self.dependencies.add(feature_alias)


class AliasVisitor(NodeVisitor):
    def __init__(self) -> None:
        self.alias_to_original = dict[str, str]()

    def visit_ImportFrom(self, node: ImportFrom) -> Any:
        match node:  # noqa
            case ImportFrom(
                names=[alias(name=name, asname=str(asname))],
            ):
                self.alias_to_original[name] = asname


def find_deps_raw(feature: AnyFeature) -> set[str]:
    final_deps = set[str]()

    source_file = getsourcefile(feature.resolver)
    source_function = getsource(feature.resolver)

    if not source_file:
        raise Exception("No source file? Strange stuff.")

    source_ast = get_source_file_ast(Path(source_file))

    alias_visitor = AliasVisitor()
    alias_visitor.visit(source_ast)

    dependency_visitor = DependencyVisitor()
    dependency_visitor.visit(ast.parse(source_function))

    for feature_alias in dependency_visitor.dependencies:
        original_name = alias_visitor.alias_to_original.get(
            feature_alias, feature_alias
        )
        final_deps.add(original_name)

    return final_deps


def find_deps(
    feature: AnyFeature, available_features: Mapping[str, AnyFeature]
) -> set[AnyFeature]:
    raw_deps = find_deps_raw(feature)
    return set(available_features[f] for f in raw_deps)
